from typing import Dict, TypeVar, Generic, Tuple
from weakref import ref

V = TypeVar('V')


class WeakTupleDict(Generic[V]):
    def __init__(self):
        self.inner: Dict[Tuple[ref], V] = {}

    def __getitem__(self, item: tuple):
        t = tuple(ref(i) for i in item)
        return self.inner[t]

    def get(self, item: tuple, default=None):
        t = tuple(ref(i) for i in item)
        return self.inner.get(t, default)

    def __setitem__(self, key, value):
        def del_callback(r):
            nonlocal t
            self.inner.pop(t, None)
        t = tuple(ref(i, del_callback) for i in key)
        self.inner[t] = value

    def __len__(self):
        return len(self.inner)



from __future__ import annotations

from abc import get_cache_token
from collections import defaultdict, ChainMap
from functools import partial
from inspect import signature, Parameter
from typing import Callable, TypeVar, Generic, Dict, Set, List, Mapping, get_type_hints, Union, Tuple, Optional, \
    MutableMapping
from weakref import WeakValueDictionary, proxy


from dyndis.annotation_filter import AnnotationFilter, annotation_filter
from dyndis.exceptions import AmbiguityError
from dyndis.implementor import Implementor
from dyndis.topological_sort import topological_sort
from dyndis.weaktupledict import WeakTupleDict

T = TypeVar('T')


class Candidate(Generic[T]):
    def __init__(self, callback: Callable[..., T], filters: Tuple[AnnotationFilter], owner: MultiDispatch, *,
                 initial_definitions: Optional[Dict[TypeVar, AnnotationFilter]] = None):
        self.callback = callback
        self.filters = filters
        self.owner = proxy(owner)
        self.initial_definitions = initial_definitions or {}

    def __str__(self):
        return self.owner.__name__ + "<" + ", ".join(str(f) for f in self.filters) + ">"

    def envelops(self, other: Candidate):
        partial = False
        for a, b in zip(self.filters, other.filters):
            if a.envelops(b):
                if b.envelops(a):
                    # a and b are equivalent
                    continue
                partial = True
            else:
                return False
        return partial

    def match(self, args):
        defined = dict(self.initial_definitions)
        for a, f in zip(args, self.filters):
            r = f.match(a, defined)
            if not r:
                return False
            if isinstance(r, Mapping):
                defined.update(r)
        return True


LookupLayer = Union[Exception, Candidate]


class MultiDispatch(Generic[T], Callable[..., T]):
    _Implementors: MutableMapping[str, Implementor] = WeakValueDictionary()

    def __init__(self, default_callback: Callable[..., T]):
        self.default_callback = default_callback
        self.__name__ = default_callback.__name__
        self.candidate_sets: Dict[int, Set[Candidate]] = defaultdict(set)

        self._cache_token = get_cache_token()
        self._layers_cache: Dict[int, List[Set[Candidate]]] = {}
        self._lookup_cache: Dict[int, WeakTupleDict[List[LookupLayer]]] = defaultdict(WeakTupleDict)

    def _add_candidate(self, func, filters, **kwargs):
        cand = Candidate(func, filters, self, **kwargs)
        self.candidate_sets[len(filters)].add(cand)

        self.clear_cache(len(filters))

    def clear_cache(self, arg_len=None):
        if arg_len is None:
            self._layers_cache.clear()
            self._lookup_cache.clear()
        else:
            self._layers_cache.pop(arg_len, None)
            self._lookup_cache.pop(arg_len, None)

    def _topological_candidates(self, func_len) -> List[Set[Candidate]]:
        if func_len in self._layers_cache:
            return self._layers_cache[func_len]
        ret = self._layers_cache[func_len] = list(topological_sort(self.candidate_sets[func_len]))
        return ret

    def _get_lookup_layers(self, t_args: Tuple[type, ...]) -> List[LookupLayer]:
        cached_lookup = self._lookup_cache[len(t_args)].get(t_args)
        if cached_lookup is not None:
            return cached_lookup
        ret = []
        tc = self._topological_candidates(len(t_args))
        for layer in tc:
            try:
                valid_cands = [c for c in layer if c.match(t_args)]
            except TypeError as e:
                ret.append(e)
                break
            if len(valid_cands) == 1:
                ret.append(valid_cands[0])
            elif valid_cands:
                ret.append(AmbiguityError(f'ambiguous call between {", ".join(str(v) for v in valid_cands)}'))
                break
        self._lookup_cache[len(t_args)][t_args] = ret
        return ret

    def __call__(self, *args, **kwargs):
        new_cache_token = get_cache_token()
        if new_cache_token != self._cache_token:
            # our cache is out of date and must be rebuilt
            self.clear_cache()
            self._cache_token = new_cache_token

        t_args = tuple(type(a) for a in args)
        ll = self._get_lookup_layers(t_args)
        for layer in ll:
            if isinstance(layer, Exception):
                raise layer
            ret = layer.callback(*args, **kwargs)
            if ret is not NotImplemented:
                return ret
        return self.default_callback(*args, **kwargs)

    def register(self, func=None, extra_namespace=None, default_annotations=None, **kwargs):
        if not func:
            return partial(self.register, **kwargs)

        sign = signature(func)
        type_hints = get_type_hints(func, localns=extra_namespace)
        if default_annotations:
            type_hints = ChainMap(default_annotations, type_hints)
        filters = []
        for p in sign.parameters.values():
            if p.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD, Parameter.POSITIONAL_ONLY):
                break
            if p.name not in type_hints:
                raise TypeError(f'parameter {p.name} is not annotated')
            filter_ = annotation_filter(type_hints[p.name])
            filters.append(filter_)
        self._add_candidate(func, tuple(filters), **kwargs)
        return func

    def implement(self, key, **kwargs):
        implementor = self._Implementors.get(key)
        if not implementor:
            implementor = self._Implementors[key] = Implementor()
        return partial(implementor, self, kwargs)

    def __get__(self, instance, owner):
        if instance:
            return partial(self, instance)
        return self



from collections import defaultdict
from typing import TypeVar, Iterable, Iterator, Set

T = TypeVar('T')


def topological_sort(members: Iterable[T]) -> Iterator[Set[T]]:
    dependencies = {}
    dependants = defaultdict(set)
    next_layer = set()
    for a in members:
        d = set()
        for b in members:
            if a.envelops(b):
                d.add(b)
                dependants[b].add(a)
        if d:
            dependencies[a] = d
        else:
            next_layer.add(a)

    yield next_layer

    while dependencies:
        new_layer = set()
        for done in next_layer:
            for freed in dependants[done]:
                dependencies[freed].remove(done)
                if not dependencies[freed]:
                    dependencies.pop(freed)
                    new_layer.add(freed)
        if not new_layer:
            raise RuntimeError('cycle')
        next_layer = new_layer
        yield next_layer



from collections import ChainMap
from functools import update_wrapper
from types import MethodDescriptorType
from typing import Callable, List, NamedTuple, Dict, Any, Optional
from warnings import warn


class PendingImplementation(NamedTuple):
    func: Callable
    md: Any
    kwargs: Optional[Dict[str, Any]]

    def chain_kwargs(self, **parents: ChainMap):
        if not self.kwargs:
            return parents

        ret = dict(self.kwargs)
        for k, parent in parents.items():
            child = ret.get(k)
            if not child:
                ret[k] = parent
            else:
                ret[k] = parent.new_child(child)
        return ret


class Implementor:
    def __init__(self):
        self.queue: List[PendingImplementation] = []
        self.done = False

    def commit(self, cls: type):
        if self.done:
            return
        self.done = True
        default_annotations = ChainMap({'self': cls, 'cls': type(cls)})
        extra_namespace = ChainMap({cls.__name__: cls})
        for queued in self.queue:
            kwargs = queued.chain_kwargs(
                default_annotations=default_annotations,
                extra_namespace=extra_namespace,
            )
            queued.md.register(
                queued.func,
                **kwargs
            )

    def __call__(self, md, kwargs, func):
        if isinstance(func, classmethod):
            func = func.__func__

        self.queue.append(PendingImplementation(func, md, kwargs))
        return ImplementorWrapper(self, func)

    def __del__(self):
        if not self.done:
            warn(RuntimeWarning('implementor deleted without registering implementations'))


class ImplementorWrapper:
    def __init__(self, implementor: Implementor, func: MethodDescriptorType):
        self.implementor = implementor
        self.inner = func
        update_wrapper(self, func)

    def __set_name__(self, owner, name):
        self.implementor.commit(owner)
        self.implementor = None
        setattr(owner, name, self.inner)

    # although the wrapper does its best to self-destruct, it might still be held as the function, so we have it
    # imitate its inner value as much as we can

    def __get__(self, instance, owner):
        return self.inner.__get__(instance, owner)

    def __call__(self, *args, **kwargs):
        return self.inner(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self.inner, item)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Mapping, TypeVar, Union, FrozenSet, Any

try:
    from typing import TypedDict
except ImportError:
    TYPED_DICT_META = None
else:
    TYPED_DICT_META = type(TypedDict('blank', {}))
    del TypedDict


def is_type_like(t):
    try:
        isinstance(None, t)
    except TypeError:
        return False
    return True


def annotation_filter(x):
    if TYPED_DICT_META and isinstance(x, TYPED_DICT_META):
        raise TypeError('cannot use a Typed Dict as a multidispatch annotation')

    if is_type_like(x):
        return ClassAnnotationFilter(x)
    if x in (None, ..., NotImplemented):
        return annotation_filter(type(x))
    if x is Any:
        return AnyAnnotationFilter
    if isinstance(x, TypeVar):
        if x.__constraints__:
            return ConstrainedTypeVarAnnotatedFilter(x)
        return BoundedTypeVarAnnotatedFilter(x)

    origin = getattr(x, '__origin__', None)
    if origin is Union:
        args = frozenset(annotation_filter(a) for a in x.__args__)
        return UnionAnnotationFilter(args)

    raise TypeError(x)


class AnnotationFilter(ABC):
    @abstractmethod
    def match(self, x: type, defined: Mapping[TypeVar, AnnotationFilter]) \
            -> Union[bool, Mapping[TypeVar, AnnotationFilter]]:
        pass

    @abstractmethod
    def envelops(self, other: AnnotationFilter) -> bool:
        """
        `self` envelops `other` if, for any defined mapping `d` and type 't',
         `self.match(t, d) or not other.match(t, d)`
        """
        return other.enveloped_by(self)

    def enveloped_by(self, other: AnnotationFilter) -> bool:
        raise TypeError

    @abstractmethod
    def __str__(self):
        pass


class ClassAnnotationFilter(AnnotationFilter):
    def __init__(self, cls: type):
        self.cls = cls

    def match(self, x, defined: Mapping[TypeVar, AnnotationFilter]):
        return issubclass(x, self.cls)

    def envelops(self, other: AnnotationFilter) -> bool:
        if isinstance(other, ClassAnnotationFilter):
            return issubclass(other.cls, self.cls)
        return super().envelops(other)

    def __str__(self):
        return self.cls.__name__

    def __eq__(self, other):
        return type(self) == type(other) and self.cls == other.cls

    def __hash__(self):
        return hash(self.cls)


class UnionAnnotationFilter(AnnotationFilter):
    def __init__(self, args: FrozenSet[AnnotationFilter]):
        self.args = args

    def match(self, x, defined: Mapping[TypeVar, AnnotationFilter]):
        results = (a.match(x, defined) for a in self.args)
        first_truish = False
        for r in results:
            if r is True:
                if first_truish is True:
                    continue
                elif not first_truish:
                    first_truish = True
                else:
                    raise TypeError(f'cannot match {x} to Union due to conflicting results')
            elif r:
                if first_truish:
                    raise TypeError(f'cannot match {x} to Union due to conflicting results')
                first_truish = r
        return first_truish

    def envelops(self, other: AnnotationFilter) -> bool:
        if isinstance(other, ClassAnnotationFilter):
            return any(a.envelops(other) for a in self.args)
        if isinstance(other, UnionAnnotationFilter):
            return all(self.envelops(a) for a in other.args)
        return super().envelops(other)

    def enveloped_by(self, other: AnnotationFilter) -> bool:
        if isinstance(other, ClassAnnotationFilter):
            return all(other.envelops(a) for a in self.args)
        return super().enveloped_by(other)

    def __eq__(self, other):
        return type(self) == type(other) and self.args == other.args

    def __hash__(self):
        return hash(self.args)

    def __str__(self):
        return '(' + "|".join(str(i) for i in self.args) + ")"


class TypeVarAnnotationFilter(AnnotationFilter):
    def __init__(self, tv: TypeVar):
        self.tv = tv

    @abstractmethod
    def match(self, x, defined: Mapping[TypeVar, AnnotationFilter]) -> Union[bool, Mapping[TypeVar, AnnotationFilter]]:
        already_defined = defined.get(self.tv)
        if already_defined:
            return already_defined.match(x, defined)

    def __eq__(self, other):
        return type(self) == type(other) and self.tv == other.tv

    def __hash__(self):
        return hash(self.tv)

    def __str__(self):
        return '~' + self.tv.__name__


class ConstrainedTypeVarAnnotatedFilter(TypeVarAnnotationFilter):
    def __init__(self, tv: TypeVar):
        super().__init__(tv)
        self.constraints = [annotation_filter(c) for c in tv.__constraints__]

    def match(self, x, defined: Mapping[TypeVar, AnnotationFilter]) -> Union[bool, Mapping[TypeVar, AnnotationFilter]]:
        s = super().match(x, defined)
        if s is not None:
            return s
        matches = list(filter(lambda t: t[1], ((af, af.match(x, defined)) for af in self.constraints)))
        if not matches:
            return False
        if len(matches) == 1:
            af, match = matches[0]
            if match is True:
                return {self.tv: af}
            return {**match, self.tv: af}
        raise TypeError(f'ambiguous typevar match, value {x} matched {", ".join(a for (a, _) in matches)}')

    def envelops(self, other: AnnotationFilter) -> bool:
        return self == other or all(c.envelops(other) for c in self.constraints)

    def enveloped_by(self, other: AnnotationFilter) -> bool:
        return self == other or all(other.envelops(c) for c in self.constraints)


class BoundedTypeVarAnnotatedFilter(TypeVarAnnotationFilter):
    def __init__(self, tv: TypeVar):
        super().__init__(tv)
        self.bound = annotation_filter(tv.__bound__ or object)

    def match(self, x, defined: Mapping[TypeVar, AnnotationFilter]) -> Union[bool, Mapping[TypeVar, AnnotationFilter]]:
        s = super().match(x, defined)
        if s is not None:
            return s
        sub_match = self.bound.match(x, defined)
        if sub_match is True:
            return {self.tv: annotation_filter(x)}
        elif sub_match:
            return {**sub_match, self.tv: annotation_filter(x)}
        return False

    def envelops(self, other: AnnotationFilter) -> bool:
        return self == other

    def enveloped_by(self, other: AnnotationFilter) -> bool:
        return self == other or other.envelops(self.bound)


class _AnyAnnotationFilter(AnnotationFilter):
    def match(self, x: type, defined: Mapping[TypeVar, AnnotationFilter]):
        return True

    def envelops(self, other: AnnotationFilter) -> bool:
        return True

    def enveloped_by(self, other: AnnotationFilter) -> bool:
        return other is self

    def __str__(self):
        return 'Any'


AnyAnnotationFilter = _AnyAnnotationFilter()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dyndis']

package_data = \
{'': ['*']}

install_requires = \
['sortedcontainers']

setup_kwargs = {
    'name': 'dyndis',
    'version': '0.2.0',
    'description': '',
    'long_description': '# Dyndis\n\n[`pip install dyndis`](https://pypi.org/project/dyndis/)\n\n## About\n\nDyndis is a library to easily and fluently make multiple-dispatch functions and methods. It was originally made for\noperators in non-strict hierarchical systems but can also serve any other multiple-dispatch purpose.\n\n## Simple Example\n\n```python\nfrom typing import Union\n\nfrom dyndis import MultiDispatch\n\n\n@MultiDispatch\ndef foo(a, b):\n    # default implementation in case no candidates match\n    raise TypeError\n\n\n@foo.register()\ndef _(a: int, b: Union[int, str]):\n    return "overload 1 <int, (int|str)>"\n\n\n@foo.register()\ndef _(a: object, b: float):\n    return "overload 2 <any, float>"\n\n\nfoo(1, "hello")  # overload 1\nfoo(("any", "object", "here"), 2.5)  # overload 2\nfoo(2, 3)  # overload 1\nfoo(2, 3.0)  # overload 2\n```\n\n## Features\n\n* dynamic upcasting.\n* seamless usage of type-hints and type variables.\n* advanced data structures to minimize candidate lookup time.\n* implementor interface makes it easy to create method-like overloads\n\n## How Does it Work?\n\nThe central class (and only one users need to import) is `MultiDispatch`. `MultiDispatch` contains candidate\nimplementations sorted by both priority and types of the parameters they accept. When the `MultiDispatch` is called, it\ncalls its relevant candidates (ordered by both priority, inheritance, and compatibility, expanded upon below) until one\nreturns a non-`NotImplemented` return value.\n\n## The Lookup Order\n\nAll candidates for parameters of types <T0, T1, T2..., TN> are ordered as follows:\n\n* Any candidate with a types that is incompatible with any type in the key is excluded. That is, if for any 0 <= `i` <=\n  N, a candidate\'s type constraint for parameter `i` is not a superclass of Ti, the candidate is excluded.\n* Candidates are ordered by inheritance. A candidate is considered to inherit another candidate if all its parameter\n  types are subclasses of (or are likewise covered by) the other\'s respective parameter type. A candidate will be\n  considered before any other candidate it inherits. So for example, <int,object> will be considered before <Number,\n  object>.\n\nIf there are two candidates that do not inherit each other, an exception (of type `dyndis.AmbiguityError`) is raised (\nunless a candidate with greater precedence than both succeeds first).\n\nIf a candidate returns `NotImplemented`, the next candidate in the order is tried. If no candidates are accepted or all\ncandidates returned `NotImplemented`, the default implementation is called.\n\n## Topology and Caches\n\n`dyndis` uses a topological set to order all its candidates by the parameter types, so that most of the candidates can\nbe disregarded without any overhead.\n\nConsidering all these candidates for every lookup gets quite slow and encumbering very quickly. For this reason,\nevery `MultiDispatch` automatically caches these computation for both sorting and processing candidates.\n\n## Default, Variadic, and Keyword parameters\n\n* If a candidate has positional parameters with a default value and a type annotation, the default value will be ignored\n  for the purposes of candidate resolution.\n* If a candidate has a variadic positional parameter, it is ignored. When called from a `MultiDispatch`, its value will\n  always be `()`.\n* If a candidate has keyword-only parameters, the parameter will not be considered for candidate types, it must either\n  have a default value or be set when the `MultiDispatch` is called.\n* If a candidate has a variadic keyword parameter, it is ignored. When called from a `MultiDispatch`, its value will be\n  according to the (type-ignored) keyword arguments.\n\nIn general, when a `MultiDispatch` is called with keyword arguments, those arguments are not considered for candidate\nresolution and are sent to each attempted candidate as-is.\n\n## Implementors\n\nan `Implementor` is a descriptor that makes it easy to create method-like candidates inside classes.\n\n```python\nfrom dyndis import MultiDispatch\n\n\n@MultiDispatch\ndef add(self, other):\n    return NotImplemented\n\n\nclass Base:\n    __add__ = add\n\n\nclass A(Base):\n    @add.implement(__qualname__)\n    def add(self, other: \'A\'):\n        # in implementor methods, `self` is assumed to be of the owner class\n        return "A+A"\n\n    @add.implement(__qualname__)\n    def add(self, other: Base):\n        return "A+Base"\n\n\nclass B(Base):\n    @add.implement(__qualname__)\n    def add(self, other: A):\n        return \'B+A\'\n\n\na = A()\nb = B()\nbase = Base()\na + b  # A+B\na + base  # A+Base\na + a  # A+A\n```\n\n## Special Type Annotations\n\ntype annotations can be of any type, or among any of these special values\n\n* `typing.Union`: accepts parameters of any of the enclosed type\n* `typing.Optional`: accepts the enclosed type or `None`\n* `typing.Any`: is considered a supertype for any type, including `object`\n* Any of typing\'s aliases and abstract classes such as `typing.List` or `typing.Sized`: equivalent to their origin\n  type (note that specialized aliases such as `typing.List[str]` are invalid)\n* `typing.TypeVar`: see below\n* `None`, `...`, `NotImplemented`: equivalent to their types\n\n## `TypeVar` annotations\n\nParameters can also be annotated with `typing.TypeVar`s. These variables bind greedily as they are encountered, and\ncount as matched upon first binding. After first binding, they are treated as the bound type (or the lowest constraint\nof the `TypeVar`) for all respects.\n\n```python\nfrom typing import TypeVar, Any\n\nfrom dyndis import MultiDispatch\n\nT = TypeVar(\'T\')\n\n\n@MultiDispatch\ndef foo(*args):\n    raise TypeError\n\n\n@foo.register()\ndef _(a: T, b: T):\n    return "type(b) <= type(a)"\n\n\n@foo.register()\ndef _(a: Any, b: Any):\n    return "type(b) </= type(a"\n\n\nfoo(1, 1)  # <=\nfoo(1, True)  # <=\nfoo(2, \'a\')  # </=\nfoo(object(), object())  # <=\n# type variables bind greedily, meaning their exact value will be equal to the first type they encounter\nfoo(False, 2)  # </=\n```\n',
    'author': 'ben avrahami',
    'author_email': 'dontcallme@illcall.you',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bentheiii/dyndis',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
